from collections.abc import Generator
# from dataclasses import dataclass, field
from pathlib import Path

import ulid
from pydantic import BaseModel, Field, ConfigDict

from cx_studio.utils import FunctionalUtils
from cx_studio.utils import PathUtils
from cx_wealth import rich_types as r
from .argument_group import ArgumentGroup


# @dataclass(frozen=True)
class Mission(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

    mission_id: ulid.ULID = Field(default_factory=ulid.new, kw_only=True)

    preset_id: str
    preset_name: str
    ffmpeg: str

    source: Path
    standard_target: Path
    overwrite: bool = False
    hardware_accelerate: str = "auto"
    options: ArgumentGroup = Field(default_factory=ArgumentGroup)
    inputs: list[ArgumentGroup] = []
    outputs: list[ArgumentGroup] = []

    @property
    def name(self):
        return PathUtils.get_basename(self.source)

    def __rich__(self):
        return r.Text.assemble(
            *[
                r.Text.from_markup(x)
                for x in FunctionalUtils.iter_with_separator(self.__rich_label__(), " ")
            ],
            overflow="crop",
        )

    def __rich_label__(self):
        yield "[bold bright_black]M[/]"
        yield f"[dim green][[cyan]{self.preset_name}[/cyan]:{len(self.inputs)}->{len(self.outputs)}][/dim green]"
        yield f"[yellow]{self.name}[/]"
        yield f"[italic dim blue]({self.source.resolve().parent})[/]"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Mission):
            return False
        return self.source == value.source and self.preset_id == value.preset_id

    def __hash__(self) -> int:
        return hash(str(self.source)) ^ hash(self.preset_id) ^ hash("mission")

    def iter_arguments(
        self,
        force_overwrite: bool | None = None,
        quote_mode: PathUtils.PathQuoteMode = "none",
    ) -> Generator[str]:
        if self.hardware_accelerate:
            yield "-hwaccel"
            yield self.hardware_accelerate
        overwrite = self.overwrite if force_overwrite is None else force_overwrite
        yield "-y" if overwrite else "-n"
        yield from self.options.iter_arguments()
        for input_group in self.inputs:
            yield from input_group.iter_arguments()
            yield "-i"
            yield PathUtils.quote(input_group.filename, quote_mode)
        for output_group in self.outputs:
            yield from output_group.iter_arguments()
            yield PathUtils.quote(output_group.filename, quote_mode)

    def __rich_detail__(self):
        yield "名称", self.name
        yield "来源预设", f"{self.preset_name}({self.preset_name})"
        yield "来源文件路径", self.source
        yield "标准目标路径", self.standard_target
        yield "覆盖已存在的目标", "是" if self.overwrite else "否"
        yield "硬件加速模式", self.hardware_accelerate
        if self.options:
            yield "通用参数（自定义）", r.Columns(
                self.options.iter_arguments(position_for_position_arguments="front")
            )
        yield "媒体输入组", self.inputs
        yield "媒体输出组", self.outputs

        yield "命令参数预览", " ".join(
            ["(ffmpeg)"] + list(self.iter_arguments(quote_mode="force"))
        )

    def iter_output_filenames(self) -> Generator[Path]:
        for output_group in self.outputs:
            if output_group.filename is not None:
                yield output_group.filename

    def iter_input_filenames(self) -> Generator[Path]:
        for input_group in self.inputs:
            if input_group.filename is not None:
                yield input_group.filename



from typing import override
from cx_studio.core import DataPackage
from cx_tools.app import IAppEnvironment
from rich.console import Console
from .arg_parser import AppContext


class AppEnv(IAppEnvironment):
    def __init__(self):
        super().__init__()
        self.app_name = "MediaScout"
        self.app_version = "0.1.0"
        self.output_console = Console()
        self.context: AppContext

    def print(self, *args, **kwargs):
        self.output_console.print(*args, **kwargs)

    def start(self):
        self.context = AppContext.load()

    def stop(self):
        pass

    def is_debug_mode_on(self):
        return self.context.debug_mode

    def show_banner(self):
        self.say("[blue]{}[/] [yellow]v{}[/]".format(self.app_name, self.app_version))


appenv = AppEnv()



from collections import defaultdict
from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Literal

from rich.columns import Columns


class ArgumentGroup:
    def __init__(
        self,
        options: dict | str | list | None = None,
        filename: Path | None = None,
        **kwargs,
    ):
        self.filename: Path | None = filename
        self._options: dict[str, list[str]] = defaultdict(list)
        self._position_arguments: list[str] = []
        if options:
            self.add_options(options)
        self.add_options(**kwargs)

    @staticmethod
    def __iter_pairs_from_list(args: list) -> Generator[tuple[str | None, str | None]]:
        prev = None
        for x in args:
            x = str(x)
            if x.startswith("-"):
                # if prev:
                yield x, None
                prev = x
            else:
                yield prev, x
                prev = None

    def __add_options_from_pairs(self, pairs: Iterable[tuple[str | None, str | None]]):
        for k, v in pairs:
            # print(k, v)
            k = self._clean_up_key(k)
            v = str(v) if v else None
            if k and v:
                self._options[k].append(v)
                continue

            if k and not v and k not in self._options:
                self._options[k] = []
                continue

            if not k and v:
                self._position_arguments.append(v)

    def __make_pairs(
        self, options: str | list | dict
    ) -> Iterable[tuple[str | None, str | None]]:
        pairs = []
        if isinstance(options, dict):
            pairs = options.items()
        elif isinstance(options, list):
            pairs = self.__iter_pairs_from_list(options)
        elif isinstance(options, str):
            pairs = self.__iter_pairs_from_list(options.split(" "))
        return pairs

    def add_options(
        self, options: str | list | dict | None = None, *args, **kwargs
    ) -> "ArgumentGroup":
        pair_iterators = []
        if options:
            pair_iterators.append(self.__make_pairs(options))

        if args:
            pair_iterators.append(self.__make_pairs(list(args)))

        if kwargs:
            pair_iterators.append(self.__make_pairs(kwargs))

        for pairs in pair_iterators:
            self.__add_options_from_pairs(pairs)
        return self

    @staticmethod
    def _format_key(key: str) -> str:
        return key if key.startswith("-") else f"-{key}"

    @staticmethod
    def _clean_up_key(key: object | None) -> str | None:
        if key is None:
            return None
        key = str(key)
        result = key[1:] if key.startswith("-") else key
        return result if len(result) > 0 else None

    def items(self) -> Generator[tuple[str, list]]:
        for k, v in self._options.items():
            yield self._format_key(k), v

    def iter_arguments(
        self,
        position_for_position_arguments: Literal[
            "front", "back", "nowhere"
        ] = "nowhere",
    ) -> Generator[str]:
        if position_for_position_arguments == "front":
            yield from self._position_arguments

        for k, v in self._options.items():
            key = self._format_key(k)
            if not v:
                yield key
            else:
                for value in v:
                    yield key
                    yield value

        if position_for_position_arguments == "back":
            yield from self._position_arguments

    def __rich_repr__(self):
        if self._position_arguments:
            yield "位置参数", Columns(self._position_arguments)
        if self._options:
            for k, v in self.items():
                yield k, Columns(v)
        if self.filename:
            yield "文件名", self.filename



import asyncio
from collections.abc import Sequence
from pathlib import Path

from cx_studio.utils import PathUtils
from jpegger.components.format_database import FormatDB
from jpegger.components.mission import Mission
from jpegger.filters import ImageFilterChain
from .simple_appcontext import SimpleAppContext


class SimpleMissionBuilder:
    DEFAULT_SAVING_OPTIONS: dict = {"saveall": True}

    def __init__(
        self,
        filter_chain: ImageFilterChain,
        app_context: SimpleAppContext,
    ):
        self.filter_chain = filter_chain
        self.output_dir = PathUtils.normalize_path(app_context.output_dir or Path.cwd())

        target_format = app_context.format
        self.target_format_info = (
            FormatDB.search(target_format) if target_format else None
        )
        self.target_suffix = None
        if self.target_format_info:
            self.target_suffix = PathUtils.normalize_suffix(target_format or "").lower()
            if self.target_suffix not in self.target_format_info.extensions:
                self.target_suffix = self.target_format_info.preferred_extension

        self.quality = app_context.quality

        self._semaphore = asyncio.Semaphore(10)

    async def make_mission(self, source: Path | str) -> Mission:
        async with self._semaphore:
            source = PathUtils.normalize_path(source)
            target = self.output_dir / source.name
            if self.target_suffix:
                target = PathUtils.force_suffix(target, self.target_suffix)

            options = self.DEFAULT_SAVING_OPTIONS.copy()
            if self.quality:
                options["quality"] = self.quality

            return Mission(
                source=source,
                target=target,
                target_format=(
                    self.target_format_info.name if self.target_format_info else None
                ),
                filter_chain=self.filter_chain,
                saving_options=options,
            )

    async def _dispatch_missions(self, sources: Sequence[Path | str]):
        missions = await asyncio.gather(*[self.make_mission(s) for s in sources])
        return [m for m in missions if m is not None]

    def make_missions(self, sources: Sequence[Path | str]) -> list[Mission]:
        return asyncio.run(self._dispatch_missions(sources))



import asyncio
import itertools
import threading
from collections.abc import Sequence, Generator, Iterable
from pathlib import Path

from cx_tools.app import ProgressTaskAgent
from cx_wealth import WealthLabel, IndexedListPanel
from cx_wealth import rich_types as r
from .argument_group import ArgumentGroup
from .mission import Mission
from .preset import Preset
from .preset_tag_replacer import PresetTagReplacer
from .source_expander import SourceExpander
from ..appenv import appenv


class MissionMaker:
    _lock = threading.Lock()

    def __init__(self, preset: Preset):
        self._preset = preset
        self._source_expander = SourceExpander(self._preset)

    def make_mission(self, source: Path, external_output_dir: Path | None) -> Mission:
        replacer = PresetTagReplacer(
            self._preset, source, output_dir=external_output_dir
        )

        general = ArgumentGroup()
        general.add_options(list(replacer.read_value_as_list(self._preset.options)))

        inputs = []
        for g in self._preset.inputs:
            x = ArgumentGroup()
            x.filename = Path(replacer.read_value(g.filename))
            x.add_options(list(replacer.read_value_as_list(g.options)))
            inputs.append(x)

        outputs = []
        for g in self._preset.outputs:
            x = ArgumentGroup()
            x.filename = Path(replacer.read_value(g.filename))
            x.add_options(list(replacer.read_value_as_list(g.options)))
            outputs.append(x)

        _overwrite: bool = self._preset.overwrite
        if appenv.context.force_overwrite:
            _overwrite = True
        if appenv.context.force_no_overwrite:
            _overwrite = False

        return Mission(
            preset_id=self._preset.id,
            preset_name=self._preset.name,
            ffmpeg=self._preset.ffmpeg,
            source=source,
            standard_target=replacer.standard_target,
            overwrite=_overwrite,
            hardware_accelerate=self._preset.hardware_accelerate or "auto",
            options=general,
            inputs=inputs,
            outputs=outputs,
        )

    def expand_sources(self, sources: Iterable[str | Path]) -> Generator[Path]:
        yield from self._source_expander.expand(*sources)

    def report(self, missions: list):
        with self._lock:
            appenv.whisper(
                IndexedListPanel(
                    missions,
                    title="预设 [red]{}[/red] 生成的任务列表".format(self._preset.name),
                )
            )

            count = len(missions)
            preset_label = WealthLabel(self._preset, justify="left", overflow="crop")
            missions_label = r.Text(f"{count}个任务", style="italic", justify="right")
            appenv.say(r.Columns([preset_label, missions_label], expand=True))

    def expand_and_make_missions(
        self, sources: Sequence[str | Path], external_output_dir: Path | None = None
    ) -> Generator[Mission]:
        wanna_quit = False
        for source in sources:
            source = Path(source)
            for ss in self._source_expander.expand(source):
                if wanna_quit:
                    break
                if appenv.wanna_quit_event.is_set():
                    wanna_quit = True
                    appenv.wanna_quit_event.clear()

                m = self.make_mission(ss, external_output_dir=external_output_dir)
                appenv.pretending_sleep(0.05)
                yield m

    @staticmethod
    async def auto_make_missions(
        presets: Iterable[Preset],
        sources: Iterable[str | Path],
        external_output_dir: Path | str | None = None,
    ) -> list[Mission]:
        # missions = []

        async def work(
            _preset: Preset, _sources: Iterable[str | Path]
        ) -> list[Mission]:
            external_dir = (
                Path(external_output_dir).resolve() if external_output_dir else None
            )
            result = []
            appenv.whisper("开始为预设<{}>扫描源文件并创建任务…".format(_preset.name))
            async with ProgressTaskAgent(
                appenv.progress, task_name=_preset.name
            ) as task_agent:
                maker = MissionMaker(_preset)
                expanded_sources = list(maker.expand_sources(_sources))
                task_agent.set_total(len(expanded_sources))
                task_agent.start()
                for s in expanded_sources:
                    wanna_quit = False
                    if appenv.really_wanna_quit_event.is_set():
                        wanna_quit = True
                        appenv.really_wanna_quit_event.clear()
                    if wanna_quit:
                        appenv.say(
                            "用户中断，[red]未为预设[cyan]{}[/]生成全部任务[/red]".format(
                                _preset.name
                            )
                        )
                        break
                    m = maker.make_mission(Path(s), external_dir)
                    result.append(m)
                    task_agent.advance()
                    await appenv.pretending_asleep(0.05)
                await appenv.pretending_asleep(0.2)
                return result

        tasks = []
        for preset in presets:
            task = asyncio.create_task(work(preset, sources))
            tasks.append(task)
            await appenv.pretending_asleep(0.2)

        results = await asyncio.gather(*tasks)
        missions = list(itertools.chain(*results))

        return missions



import os
from collections.abc import Generator
from pathlib import Path

from cx_studio.tag_replacer import TagReplacer, PathInfoProvider
from cx_studio.utils import PathUtils
from .preset import Preset


class PresetTagReplacer:
    def __init__(self, preset: Preset, source: Path, output_dir: Path | None = None):
        self._preset = preset
        self._source = Path(source)
        self.replacer = TagReplacer()
        self.replacer.install_provider("preset", self._provide_preset_info)
        self.replacer.install_provider("profile", self._provide_preset_info)
        self.replacer.install_provider("custom", self._provide_custom_values)
        self.replacer.install_provider("source", PathInfoProvider(self._source))
        self.replacer.install_provider("sep", os.sep)

        output_dir = (output_dir or Path.cwd()).resolve()
        target_folder = Path(self._preset.target_folder)
        if target_folder.is_absolute():
            output_dir = target_folder
        else:
            output_dir = Path(output_dir, target_folder)
        parent_dirs = PathUtils.get_parents(source, self._preset.keep_parent_level)
        target_name = PathUtils.force_suffix(
            PathUtils.get_basename(source), self._preset.target_suffix
        )
        self._target = Path(output_dir, *parent_dirs, target_name).resolve()
        self.replacer.install_provider("target", PathInfoProvider(self.standard_target))

    @property
    def standard_target(self) -> Path:
        return Path(self.read_value(str(self._target)))

    def _provide_preset_info(self, param: str) -> str | None:
        param = str(param).lower()
        match param:
            case "id":
                return self._preset.id
            case "name":
                return self._preset.name
            case "description":
                return self._preset.description
            case "folder":
                return str(self._preset.path.parent)
            case "folder_name":
                return self._preset.path.parent.name
            case "input_count":
                return str(len(self._preset.inputs))
            case "output_count":
                return str(len(self._preset.outputs))
        return None

    def _provide_custom_values(self, param: str) -> str | None:
        param = str(param).split(" ")[0].lower()
        result = self._preset.custom.get(param)
        return str(result) if result else None

    def read_value(self, value: str) -> str:
        return self.replacer.replace(value)

    def read_value_as_list(self, value: str | list) -> Generator[str]:
        if isinstance(value, list):
            for v in value:
                yield from self.read_value_as_list(v)
        else:
            value = self.read_value(value)
            if " " in value:
                yield from self.read_value_as_list(value.split(" "))
            else:
                yield self.read_value(value)



from pathlib import PurePath
from .inspector_info import InspectorInfo
from .media_path_inspector import MediaPathInspector
from collections.abc import Iterable
import re
import urllib.parse


class FileListInspector(MediaPathInspector):
    PATH_PATTERN = re.compile(
        r"""
        ^                           # 路径开始
        (?:                         # 非捕获组，匹配 Windows 驱动器或 Unix 根目录
            [A-Za-z]:\\?            # Windows 驱动器（可选反斜杠）
            |
            /                       # Unix/Linux/Mac 根目录
        )?
        (?:                         # 非捕获组，匹配路径中的目录和文件名
            [^\\/:*?"<>|\r\n]+      # 匹配除特殊字符外的任意字符
            (?:\\|/)                # 目录分隔符
        )*
        [^\\/:*?"<>|\r\n]*          # 最后的目录或文件名
        $                           # 路径结束
        """,
        re.VERBOSE,
    )

    URL_PATTERN = re.compile(r"file://.*/(\S+)")

    def __init__(self, *extensions: str):
        super().__init__()
        self.acceptable_extensions = set(extensions)

    def _is_inspectable(self, info: InspectorInfo) -> bool:
        if self.acceptable_extensions:
            if not info.path.suffix in self.acceptable_extensions:
                return False

        if not info.is_decodable():
            return False
        return True

    def _parse_url(self, line: str) -> str | None:
        match = self.URL_PATTERN.search(urllib.parse.unquote(line))
        return match.group(1) if match else None

    def _parse_path(self, line: str) -> str | None:
        match = self.PATH_PATTERN.search(line)
        return match.group(0) if match else None

    def _inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        with open(info.path, "r", encoding=info.encoding) as fp:
            for line in fp:
                if (url := self._parse_url(line)) is not None:
                    yield PurePath(url)
                    continue
                if (path := self._parse_path(line)) is not None:
                    yield PurePath(path)
                    continue



from typing import Literal

from PIL.Image import Image

from .image_filter import IImageFilter


class SimpleBlackWhiteFilter(IImageFilter):
    def __init__(self):
        super().__init__()

    def run(self, image: Image) -> Image:
        # Convert the image to grayscale (black and white)
        return image.convert("L").convert("RGB")


class ColorSpaceFilter(IImageFilter):
    colorspace_type = Literal["RGB", "L", "CMYK"]

    def __init__(self, colorspace: str | None):
        super().__init__()
        self.colorspace = colorspace

    def run(self, image: Image) -> Image:
        if not self.colorspace:
            return image
        if self.colorspace == "RGB" or self.colorspace == "L":
            return image.convert(self.colorspace)
        else:
            return image.convert("RGB").convert(self.colorspace)

    def __rich_label__(self):
        yield from super().__rich_label__()
        _RGB = "[black on red]R[black on green]G[black on blue]B[reset]"
        _CMYK = "[black on cyan]C[black on magenta]M[black on yellow]Y[black on black]K[reset]"
        _L = "[black on white]L[reset]"
        _NONE = "[red]N/A[/]"
        param = {
            "RGB": _RGB,
            "CMYK": _CMYK,
            "L": _L,
            None: _NONE,
        }.get(self.colorspace, _NONE)
        yield f"[blue]({param})[/]"

    def __filter_description__(self) -> str:
        return f"将图像的色彩空间转换为 {self.colorspace}"



from .config_manager import ConfigManager
from .iappenv import IAppEnvironment, CxHighlighter
from .iapplication import IApplication
from .progress_task_agent import ProgressTaskAgent
from .safe_error import SafeError



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import io
from pathlib import Path, PurePath
import os
from collections.abc import Iterable
from typing import IO, Literal, Self, override, AnyStr
from cx_studio.core import FileSize
from cx_studio.utils import EncodingUtils
from cx_studio.utils import StreamUtils

from .inspector_info import InspectorInfo


class MediaPathInspector(ABC):
    @abstractmethod
    def _is_inspectable(self, info: InspectorInfo) -> bool:
        pass

    def is_inspectable(self, info: InspectorInfo) -> bool:
        if not info.path.exists():
            return False

        return self._is_inspectable(info)

    @abstractmethod
    def _inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        pass

    def inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        try:
            yield from self._inspect(info)
        except UnicodeDecodeError:
            pass



from cx_tools.app import IAppEnvironment
from .simple_appcontext import SimpleAppContext


class AppEnv(IAppEnvironment):
    def __init__(self):
        super().__init__()
        self.app_name = "Jpegger"
        self.app_version = "0.5.1"
        self.context: SimpleAppContext

    def is_debug_mode_on(self):
        return self.context.debug_mode


appenv = AppEnv()



import asyncio
from collections.abc import Iterable, Sequence
from datetime import datetime
import itertools
import sys
from cx_studio.core.cx_time import CxTime
from cx_studio.ffmpeg import FFmpegAsync
from cx_studio.ffmpeg.cx_ff_infos import FFmpegCodingInfo
from cx_tools.app import IApplication, SafeError
from cx_wealth.indexed_list_panel import IndexedListPanel
from .appenv import appenv
from pathlib import Path
from cx_wealth import rich_types as r


class FFPrettyApp(IApplication):
    def __init__(self, arguments: Sequence[str] | None = None):
        super().__init__(arguments)
        self.arguments = []
        for a in self.sys_arguments:
            if a == "-d":
                appenv.debug_mode = True
            elif a == "--debug":
                appenv.debug_mode = True
            else:
                self.arguments.append(a)

        if "-y" not in self.arguments and "-n" not in self.arguments:
            self.arguments.append("-n")

        self.ffmpeg = FFmpegAsync(appenv.ffmpeg_executable)
        self.start_time: datetime

        self._task_description: str = ""

    def start(self):
        appenv.start()
        self.start_time = datetime.now()

    def stop(self):
        appenv.stop()
        time_span = datetime.now() - self.start_time
        if time_span.total_seconds() > 5:
            appenv.say(
                "[dim]用时 [blue]{}[/]。".format(
                    CxTime.from_seconds(time_span.total_seconds()).pretty_string
                )
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = False
        if exc_type is None:
            pass
        elif issubclass(exc_type, SafeError):
            appenv.say("[red]错误：{}[/]".format(exc_val))
            result = True
        self.stop()
        return result

    def input_files(self) -> Iterable[str]:
        input_marked = False
        for a in self.arguments:
            if a == "-i":
                input_marked = True
            elif input_marked:
                yield a
                input_marked = False

    def output_files(self) -> Iterable[str]:
        prev_key = None
        for a in self.arguments:
            if a.startswith("-"):
                prev_key = a
                continue
            if "." in a and prev_key != "-i":
                yield a

    async def _random_task_description(self):
        input_files = [Path(x).name for x in self.input_files()]
        output_files = [Path(x).name for x in self.output_files()]
        for name in itertools.cycle(input_files + output_files):
            try:
                self._task_description = name
                await asyncio.sleep(2)
            except asyncio.CancelledError:
                break

    async def execute(self, args: Iterable[str]):
        task_id = appenv.progress.add_task("[green]正在开始...[/]", total=None)

        m, n = len(list(self.input_files())), len(list(self.output_files()))
        sumary = f"[blue][{m}->{n}][/]"

        @self.ffmpeg.on("status_updated")
        def on_status_updated(status: FFmpegCodingInfo):
            current = status.current_time.total_seconds
            total = status.total_time.total_seconds if status.total_time else None

            speed = "[bright_black][{:.2f}x][/]".format(status.current_speed)
            desc = f"{sumary}{speed}[green]{self._task_description}[/]"
            appenv.progress.update(
                task_id,
                completed=current,
                total=total,
                description=desc,
            )

        ff_task = asyncio.create_task(self.ffmpeg.execute(args))
        desc_task = asyncio.create_task(self._random_task_description())

        while not ff_task.done():
            await asyncio.sleep(0.01)
            if appenv.wanna_quit_event.is_set():
                self.ffmpeg.cancel()
                break

        await asyncio.wait([ff_task])
        desc_task.cancel()
        return ff_task.result()

    def run(self) -> bool:
        if not self.arguments:
            raise SafeError("No arguments provided.")

        if not appenv.ffmpeg_executable:
            raise SafeError("No ffmpeg executable found.")

        appenv.whisper("检查输入输出……")
        inputs = list(self.input_files())
        outputs = list(self.output_files())
        appenv.whisper(
            IndexedListPanel(inputs, title="输入文件"),
            IndexedListPanel(outputs, title="输出文件"),
        )

        if not inputs:
            raise SafeError("No input files provided.")

        if not outputs:
            raise SafeError("No output files provided.")

        for i in inputs:
            if not Path(i).exists():
                raise SafeError(f"输入文件 {i} 不存在。")

        existed_outputs = [x for x in outputs if Path(x).exists()]
        if existed_outputs and "-y" not in self.arguments:
            appenv.whisper(IndexedListPanel(existed_outputs, title="已存在的输出文件"))
            raise SafeError("请使用 -y 参数覆盖已存在的文件。")

        non_exist_dirs = filter(
            lambda x: not x.exists(), [Path(a).parent for a in outputs]
        )
        if non_exist_dirs:
            for d in non_exist_dirs:
                d.mkdir(parents=True, exist_ok=True)
            appenv.whisper(IndexedListPanel(non_exist_dirs, title="创建的输出目录"))

        result = asyncio.run(self.execute(self.arguments))

        appenv.whisper("运行结果：{}".format(result))

        if not result:
            if self.ffmpeg.is_canceled:
                raise SafeError("[blue]用户取消了操作。[/]")
            else:
                raise SafeError("[red]操作失败，请排查问题。[/]")

        return result



import asyncio
from datetime import datetime
import importlib
import importlib.resources
import signal
import time
from collections.abc import Sequence
from pathlib import Path

from cx_studio.core.cx_filesize import FileSize
from cx_studio.core.cx_time import CxTime
from cx_tools.app import IAppEnvironment, ConfigManager
from cx_wealth import rich_types as r
from media_killer.components.exception import SafeError
from .appcontext import AppContext

# from .mk_help_info import MKHelp
from cx_tools import FileSizeCounter


class AppEnv(IAppEnvironment):
    def __init__(self):
        super().__init__()
        self.app_name = "MediaKiller"
        self.app_version = "0.5.1.4"
        self.app_description = "媒体文件批量操作工具"
        self.context: AppContext = AppContext()
        self.progress = r.Progress(
            # RenderableColumn("[bright_black]M[/]"),
            r.SpinnerColumn(),
            r.TextColumn(
                "[progress.description]{task.description}",
                table_column=r.Column(ratio=60, no_wrap=True),
            ),
            r.BarColumn(table_column=r.Column(ratio=40)),
            r.TaskProgressColumn(justify="right"),
            r.TimeRemainingColumn(compact=True),
            console=self.console,
            transient=True,
            expand=True,
        )

        # self.console = self.progress.console
        self.config_manager = ConfigManager(self.app_name)
        self._garbage_files = []
        self._app_start_time: datetime

        self.input_filesize_counter = FileSizeCounter()
        self.output_filesize_counter = FileSizeCounter()

    def is_debug_mode_on(self):
        return self.context.debug_mode

    def load_arguments(self, arguments: Sequence[str] | None = None):
        self.context = AppContext.from_arguments(arguments)

    def start(self):
        self.progress.start()
        self._app_start_time = datetime.now()

    def stop(self):
        self.progress.refresh()
        time.sleep(0.1)
        self.progress.stop()
        self.clean_garbage_files()
        self.config_manager.remove_old_log_files()

        input_filesize = self.input_filesize_counter.total_size
        output_filesize = self.output_filesize_counter.total_size
        filesize_report = ""
        if input_filesize.total_bytes > 0:
            filesize_report = (
                f"[dim]输入文件总大小: [cx.nummber]{input_filesize.pretty_string}[/]"
            )
        if output_filesize.total_bytes > 0:
            filesize_report += (
                f"[dim] 输出文件总大小: [cx.number]{output_filesize.pretty_string}[/]"
            )
        if len(filesize_report) > 0:
            self.say(filesize_report)

        time_spent = datetime.now() - self._app_start_time
        if time_spent.total_seconds() > 5:
            self.say(
                "[cx.whisper]总共耗时[cx.number]{}[/]。[/]".format(
                    CxTime.from_seconds(time_spent.total_seconds()).pretty_string
                )
            )

    def pretending_sleep(self, interval: float = 0.2):
        if self.context.pretending_mode:
            time.sleep(interval)

    async def pretending_asleep(self, interval: float = 0.2):
        if self.context.pretending_mode:
            await asyncio.sleep(interval)

    def add_garbage_files(self, *filenames: str | Path):
        self._garbage_files.extend(map(Path, filenames))

    def clean_garbage_files(self):
        if not self._garbage_files:
            return
        self.say("[dim]正在清理失败的目标文件...[/]")
        for filename in self._garbage_files:
            filename.unlink(missing_ok=True)
            self.whisper(f"  [cx.filepath]{filename}[/] [red]已删除[/red]")
            if self.context.debug_mode:
                time.sleep(0.1)
        if len(self._garbage_files) > 0:
            self.say(f"[dim]已清理 {len(self._garbage_files)} 个目标文件。[/]")
        else:
            self.say("[dim cx.info]没有失败的目标文件需要清理。[/]")
        self._garbage_files.clear()

    def show_banner(self):
        banners = []

        with importlib.resources.open_text("media_killer", "banner.txt") as banner:
            banner_text = r.Text(
                banner.read(),
                style="bold red",
                no_wrap=True,
                overflow="crop",
                justify="center",
            )
            banners.append(r.Align.center(banner_text))

        version_info = r.Text.from_markup(
            f"[bold blue]{self.app_name}[/] [yellow]v{self.app_version}[/]"
        )
        banners.append(r.Align.center(version_info))

        description = r.Text(self.app_description, style="bright_black")
        tags = []
        if self.context.pretending_mode:
            tags.append("[blue]模拟运行[/]")
        if self.context.force_no_overwrite:
            tags.append("[green1]安全模式[/]")
        elif self.context.force_overwrite:
            tags.append("[red]强制覆盖模式[/]")
        if tags:
            description = "·".join(tags)
        banners.append(r.Align.center(description))

        self.say(r.Group(*banners))

    def check_overwritable_file(self, filename: Path, check_only: bool = False) -> bool:
        existed = filename.exists()
        result = not existed
        if self.context.force_overwrite:
            result = True
        if self.context.force_no_overwrite:
            result = False

        if not check_only and not result:
            msg = f"文件 {filename} 已存在，"
            if self.context.force_no_overwrite:
                msg += "请取消 --force-no-overwrite 选项"
            elif not self.context.force_overwrite:
                msg += "请使用 --force-overwrite 选项尝试覆盖"
            msg += "或指定其它文件名."
            raise SafeError(msg)

        if not check_only and result:
            self.say(f"[dim red]文件 {filename} 已存在，将强制覆盖。[/]")
        return result


appenv = AppEnv()


signal.signal(signal.SIGINT, appenv.handle_interrupt)



from argparse import ArgumentParser
from collections.abc import Sequence
from typing import Literal

from rich_argparse import RichHelpFormatter


class AppContext:
    def __init__(self, **kwargs):
        self.inputs: list[str] = []
        self.script_output: str | None = None
        self.pretending_mode: bool = False
        self.debug_mode: bool = False
        self.sort_mode: Literal["source", "preset", "target", "x"] = "x"
        self.continue_mode: bool = False
        self.generate: bool = False
        self.save_script: str | None = None
        self.show_full_help: bool = False
        self.force_overwrite: bool = False
        self.force_no_overwrite: bool = False
        self.output_dir: str | None = None
        self.show_help = False
        self.max_workers: int = 1

        for k, v in kwargs.items():
            if k in self.__dict__:
                self.__dict__[k] = v

    def __rich_repr__(self):
        yield from self.__dict__.items()

    @staticmethod
    def __make_parser() -> ArgumentParser:
        parser = ArgumentParser(
            # prog="MediaKiller",
            description="MediaKiller 是一个命令行多媒体文件批量处理工具。",
            formatter_class=RichHelpFormatter,
            epilog="—— 来自 Cxalio 工作室工具集。",
            add_help=False,
        )

        parser.add_argument(
            "inputs",
            help="多个需要处理的文件路径[dim]（源文件或配置文件）",
            nargs="*",
            metavar="输入文件",
        )

        parser.add_argument(
            "-g",
            "--generate",
            help="生成新的预设文件示例",
            action="store_true",
            default=False,
            dest="generate",
        )
        parser.add_argument("--save-script", "-s", help="将转码任务编写为脚本")
        parser.add_argument(
            "-j",
            "--jobs",
            "--max-workers",
            help="指定最大工作线程数",
            type=int,
            default=1,
            dest="max_workers",
            metavar="线程数",
        )
        parser.add_argument(
            "-c",
            "--continue",
            action="store_true",
            help="重新加载上次运行的任务",
            dest="continue_mode",
        )

        parser.add_argument(
            "--output",
            "-o",
            help="指定一个输出目录",
            metavar="输出目录",
            default=None,
            dest="output_dir",
        )
        parser.add_argument(
            "--sort",
            help="指定任务排序方式",
            choices=["source", "preset", "target", "x"],
            default="x",
            metavar="排序方式代码",
            dest="sort_mode",
        )

        parser.add_argument(
            "--overwrite",
            "-y",
            help="强制覆盖所有输出文件",
            action="store_true",
            default=False,
            dest="force_overwrite",
        )

        parser.add_argument(
            "--no-overwrite",
            "-n",
            help="强制启用安全模式（不覆盖已有文件）",
            action="store_true",
            default=False,
            dest="force_no_overwrite",
        )

        parser.add_argument(
            "-h",
            "--help",
            # action="help",
            help="显示此帮助信息",
            dest="show_help",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "--tutorial",
            "--full-help",
            action="store_true",
            help="显示详细教程",
            dest="show_full_help",
        )

        parser.add_argument(
            "-p",
            "--pretend",
            help="以[italic dim]假装模式[/]模拟运行 :)",
            action="store_true",
            dest="pretending_mode",
        )
        parser.add_argument(
            "-d",
            "--debug",
            help="显示调试信息",
            action="store_true",
            dest="debug_mode",
        )

        return parser

    @classmethod
    def from_arguments(cls, arguments: Sequence[str] | None = None):
        parser = cls.__make_parser()
        args = parser.parse_args(arguments)
        return cls(**vars(args))



# from .argument_group import ArgumentGroup
# from .input_scanner import InputScanner
# from .mission import Mission
# from .mission_arranger import MissionArranger
# from .mission_maker import MissionMaker
# from .mission_master import MissionMaster
# from .mission_runner import MissionRunner
# from .preset import Preset
# from .script_maker import ScriptMaker
# from .source_expander import SourceExpander
# from .mission_xml import MissionXML



from pathlib import PurePath
from .media_path_inspector import MediaPathInspector
from collections.abc import Iterable
from typing import IO, AnyStr
import csv

from .inspector_info import InspectorInfo


class ResolveMetadataInspector(MediaPathInspector):
    DELIMITER = ","
    FILE_NAME_FIELD = "File Name"
    CLIP_DIRECTORY_FIELD = "Clip Directory"

    def _get_headers(self, info: InspectorInfo) -> list[str]:
        headers = info.peek_first_line().split(self.DELIMITER)
        return headers

    def _is_inspectable(self, info: InspectorInfo) -> bool:
        if info.path.suffix != ".csv":
            return False
        if not info.is_decodable():
            return False
        headers = self._get_headers(info)
        return (
            len(headers) > 0
            and self.FILE_NAME_FIELD in headers
            and self.CLIP_DIRECTORY_FIELD in headers
        )

    def _inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        headers = self._get_headers(info)
        with open(info.path, "r", encoding=info.encoding) as fp:
            reader = csv.DictReader(fp, fieldnames=headers, delimiter=self.DELIMITER)
            for row in reader:
                name = row.get(self.FILE_NAME_FIELD)
                folder = row.get(self.CLIP_DIRECTORY_FIELD)
                if name == self.FILE_NAME_FIELD or folder == self.CLIP_DIRECTORY_FIELD:
                    continue
                if name and folder:
                    yield PurePath(folder, name)



from typing import override

from PIL.Image import Image

from .image_filter import IImageFilter


class ImageFilterChain(IImageFilter):
    def __init__(self, filters: list):
        super().__init__()
        self.filters = filters

    def append(self, filter: IImageFilter):
        if isinstance(filter, ImageFilterChain):
            self.filters.extend(filter.filters)
        else:
            self.filters.append(filter)

    def run(self, image: Image) -> Image:
        for filter in self.filters:
            image = filter.run(image)
        return image

    @override
    def filter_name(self):
        return "FilterChain"

    def __len__(self):
        return len(self.filters)

    def __rich_label__(self):
        yield from super().__rich_label__()
        yield f"[blue]({len(self.filters)}Filters)[/]"

    def __rich_detail__(self):
        yield "Filter Chaing"
        for i, f in enumerate(self.filters):
            yield f"[dim cyan]{i} {f.filter_name()}[/]", f.__filter_description__()

    def __filter_description__(self) -> str:
        descriptions = [f.__filter_description__() for f in self.filters]
        return "；".join(descriptions)

    def step_descriptions(self) -> list[str]:
        return [f"{f.filter_name()}:{f.__filter_description__()}" for f in self.filters]

    def __rich_repr__(self):
        for i, f in enumerate(self.filters):
            yield i, f.filter_name(), f.__filter_description__()



import csv
import importlib.resources
from threading import Lock, Event

from pydantic import BaseModel, Field

from cx_studio.utils import PathUtils


class FormatInfo(BaseModel):
    name: str
    extensions: list[str]
    load_params: dict[str, str] = Field(default_factory=dict)
    save_params: dict[str, str] = Field(default_factory=dict)

    @property
    def preferred_extension(self) -> str:
        return self.extensions[0] if self.extensions else ""


class FormatDB:
    __lock = Lock()
    __default_data_loaded = Event()

    __data: dict[str, FormatInfo] = {}

    @classmethod
    def _load_default_data(cls):
        with cls.__lock:
            if cls.__default_data_loaded.is_set():
                return
            default_content = importlib.resources.read_text(
                "jpegger.components", "formats.csv"
            )

            reader = csv.DictReader(
                default_content.splitlines(), ["NAME", "EXTENSIONS"]
            )
            for row in reader:
                name = row["NAME"].strip().upper()
                extensions = [
                    PathUtils.normalize_suffix(ext.strip().lower())
                    for ext in row["EXTENSIONS"].split(" ")
                ]
                info = FormatInfo(name=name, extensions=extensions)
                cls.__data[info.name] = info
            cls.__default_data_loaded.set()

    @classmethod
    def search_for_name(cls, name: str) -> FormatInfo | None:
        return cls.__data.get(name)

    @classmethod
    def search_for_extension(cls, extension: str) -> FormatInfo | None:
        extension = PathUtils.normalize_suffix(extension).lower()
        for info in cls.__data.values():
            if extension in info.extensions:
                return info
        return None

    @classmethod
    def search(cls, keyword: str) -> FormatInfo | None:
        if result := cls.search_for_name(keyword):
            return result
        if result := cls.search_for_extension(keyword):
            return result
        return None

    @classmethod
    def formats(cls) -> list[str]:
        return list(cls.__data.keys())

    @classmethod
    def acceptable_extensions(cls) -> list[str]:
        result = {ext for x in cls.__data.values() for ext in x.extensions}
        exts = list(result)
        exts.sort()
        return exts


FormatDB._load_default_data()



from pathlib import Path

import ulid
from pydantic import BaseModel, Field, ConfigDict

from ..filters import ImageFilterChain


class Mission(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)
    mission_id: ulid.ULID = Field(default_factory=ulid.new, kw_only=True)

    source: Path
    target: Path
    target_format: str | None
    filter_chain: ImageFilterChain = ImageFilterChain([])
    saving_options: dict = {}

    def __rich_label__(self):
        yield f"[yellow][>{self.target_format or "auto"}][/]"
        yield self.source.name
        yield f"[blue]=={len(self.filter_chain)}=>[/]"
        yield self.target.name



from cx_tools.app import SafeError


class NoSourceFileError(SafeError):
    def __init__(self, message: str | None = None, style: str | None = None):
        super().__init__(message or "未找到源文件", style or "red")


class TargetingSourceFileError(SafeError):
    def __init__(self, message: str | None = None, style: str | None = None):
        super().__init__(message or "源文件无法被目标文件所覆盖", style or "red")



from typing import override
from cx_tools.app import IAppEnvironment
from cx_studio.path_expander import CmdFinder
from cx_wealth import rich_types as r
import signal


class AppEnv(IAppEnvironment):
    def __init__(self):
        super().__init__()
        self.app_name = "FFpretty"
        self.app_version = "0.1.0"
        self.ffmpeg_executable = CmdFinder.which("ffmpeg")
        self.debug_mode = False

        self.progress = r.Progress(
            # RenderableColumn("[bright_black]M[/]"),
            r.SpinnerColumn(),
            r.TextColumn(
                "[progress.description]{task.description}",
                table_column=r.Column(ratio=60, no_wrap=True),
            ),
            r.BarColumn(table_column=r.Column(ratio=40)),
            r.TaskProgressColumn(justify="right"),
            r.TimeRemainingColumn(compact=True),
            expand=True,
            console=self.console,
            transient=True,
        )

    @override
    def is_debug_mode_on(self):
        return self.debug_mode

    def start(self):
        self.whisper("FFpretty started")
        self.whisper(f"FFmpeg executable: {self.ffmpeg_executable}")
        self.progress.start()

    def stop(self):
        self.whisper("FFpretty stopped")
        self.progress.stop()


appenv = AppEnv()


signal.signal(signal.SIGINT, appenv.handle_interrupt)
